import subprocess
import xmlrpc.client

from finorch.config.config import client_config_manager
from finorch.transport.exceptions import TransportConnectionException, TransportTerminateException, \
    TransportGetJobStatusException, TransportGetJobFileException, TransportGetJobFileListException, \
    TransportGetJobsException
from finorch.transport.abstract_transport import AbstractTransport
from finorch.utils.port import test_port_open


class LocalTransport(AbstractTransport):
    """
    Transport for running jobs on the local machine (Where the API is running)
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._client_rpc = None

    def _check_client_connectivity(self, port=None):
        """
        Checks if the client is already running by reading the details from the config file and trying to connect to
        the clients last used port
        :param port: An option port to check for connectivity. If this is None the port from the client configuration
        is used.
        :return: True if the client is running and accepting connections, otherwise False
        """
        if port := port or client_config_manager.get_port():
            self._port = int(port)

            self._connected = test_port_open(self._port)

        return self._connected

    def _spawn_client(self):
        """
        Uses subprocess to spawn the client and returns the port that the client is listening on

        :return: The port that the client is listening on
        """

        # Start the client process
        args = [
            "python",
            "-m",
            "finorch.client.client",
            self._session.callsign
        ]
        p = subprocess.Popen(args, stdout=subprocess.PIPE)

        # Because we need to read the output of the client asynchronously, we need to wait for the "=EOF=" line to
        # indicate that the client process has finished with it's output.
        stdout = []
        while _line := p.stdout.readline().decode('utf-8').strip():
            if _line == "=EOF=":
                break

            stdout.append(_line)

        # Check if the client started successfully
        if stdout[0] == "error":
            # Report the error from the client
            raise TransportConnectionException('\n'.join(stdout[1:]))

        # Try to parse the first line of the output from the client as the port it is running on and check the
        # connectivity.
        self._port = str(stdout[0])
        if not self._check_client_connectivity(self._port):
            raise TransportConnectionException("Unable to connect to the port reported by the client")

    def connect(self):
        # Check if the client is already running and start it with subprocess, which will manage it's own finesse
        # processes, if it's not
        if not self._check_client_connectivity():
            self._spawn_client()

        self._client_rpc = xmlrpc.client.ServerProxy(
            f'http://localhost:{self._port}/rpc',
            allow_none=True,
            use_builtin_types=True
        )

        self._client_rpc.set_exec_path(self.exec_path)

    def start_job(self, katscript):
        return self._client_rpc.start_job(katscript)

    def terminate(self):
        if not self._connected:
            raise TransportTerminateException("Client is not connected")

        self._client_rpc.terminate()

    def disconnect(self):
        raise NotImplementedError()

    def get_job_file(self, job_identifier, file_path):
        status = self._client_rpc.get_job_file(job_identifier, file_path)
        if type(status) is bytes:
            return status
        else:
            raise TransportGetJobFileException(status[1])

    def get_job_file_list(self, job_identifier):
        status = self._client_rpc.get_job_file_list(job_identifier)
        if type(status) is list:
            return status
        else:
            raise TransportGetJobFileListException(status[1])

    def get_job_status(self, job_identifier):
        status = self._client_rpc.get_job_status(job_identifier)
        if type(status) is int:
            return status
        else:
            raise TransportGetJobStatusException(status[1])

    def get_jobs(self):
        status = self._client_rpc.get_jobs()
        if type(status) is list:
            return status
        else:
            raise TransportGetJobsException(status[1])

    def stop_job(self, job_identifier):
        raise NotImplementedError()



class JobStatus:
    # A job is pending if it is not yet submitted
    PENDING = 10
    # A job is queued if it is in the queue on the cluster it is to run on
    QUEUED = 40
    # A job is running if it is currently running on the cluster it is to run on
    RUNNING = 50
    # A job is cancelled if it was queued or running and was then cancelled
    CANCELLED = 70
    # A job is completed if it is finished running on the cluster without error
    COMPLETED = 500

    @staticmethod
    def display_name(status):
        if status == JobStatus.PENDING:
            return 'Pending'
        elif status == JobStatus.QUEUED:
            return 'Queued'
        elif status == JobStatus.RUNNING:
            return 'Running'
        elif status == JobStatus.CANCELLED:
            return 'Cancelled'
        elif status == JobStatus.COMPLETED:
            return 'Completed'
        else:
            return 'Unknown'



from finorch.sessions.cit import CitSession
from finorch.sessions.local import LocalSession
from finorch.sessions.ozstar import OzStarSession

session_map = {
    LocalSession.callsign: LocalSession,
    OzStarSession.callsign: OzStarSession,
    CitSession.callsign: CitSession
}



import datetime
import logging

from sqlalchemy import Column, Integer, String, DateTime
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from finorch.utils.job_status import JobStatus

Base = declarative_base()


class Job(Base):
    __tablename__ = 'job'

    id = Column(Integer, primary_key=True)
    batch_id = Column(Integer, unique=True, nullable=True)
    identifier = Column(String(40), unique=True)
    start_time = Column(DateTime, default=datetime.datetime.now, nullable=False)
    status = Column(Integer, default=JobStatus.PENDING)


class Database:
    def __init__(self, exec_path):
        """
        Initialises the database.

        :param exec_path: The path where the job output is kept. This is where the sqlite database will be stored.
        """
        logging.getLogger('sqlalchemy').setLevel(logging.ERROR)

        # Set up the sqlite database
        self.engine = create_engine(f"sqlite:///{exec_path / 'db.sqlite3'}")

        Base.metadata.create_all(self.engine)

        Session = sessionmaker(bind=self.engine)
        self.session = Session()

    def add_job(self, job_identifier, batch_id=None):
        """
        Inserts a new job with the specified job identifier

        :param job_identifier: The job identifier
        :return: None
        """
        job = Job(
            identifier=job_identifier,
            batch_id=batch_id
        )

        self.session.add(job)
        self.session.commit()

        return True

    def get_job_status(self, job_identifier):
        """
        Gets the status of the specified job

        :param job_identifier: The identifier of the job
        :return: The status of the job (int) if the job was found, otherwise a Tuple of (None, *reason*)
        """
        results = self.session.query(Job).filter(Job.identifier == job_identifier)

        if results.count() != 1:
            return None, f"Job with with identifier {job_identifier} not found"

        return results.first().status

    def update_job_status(self, job_identifier, new_status):
        """
        Updates the status of a specified job

        :param job_identifier: The identifier of the job
        :param new_status: The new status for the job
        :return: None
        """
        results = self.session.query(Job).filter(Job.identifier == job_identifier)

        if results.count() != 1:
            return None, f"Job with with identifier {job_identifier} not found"

        results.first().status = new_status
        self.session.commit()

        return True

    def get_jobs(self):
        """
        Gets the list of jobs

        :return: A list of dictionaries containing job information
        """
        data = [r._asdict() for r in self.session.query(Job.id, Job.identifier, Job.start_time, Job.status).all()]

        return data

    def get_job_batch_id(self, job_identifier):
        """
        Gets the batch id of the specified job

        :param job_identifier: The identifier of the job
        :return: The batch_id of the job or None
        """
        results = self.session.query(Job).filter(Job.identifier == job_identifier)

        if results.count() != 1:
            return None, f"Job with with identifier {job_identifier} not found"

        return results.first().batch_id



import logging
import os
import subprocess
import sys
import uuid
from pathlib import Path

from finorch.sessions.abstract_client import AbstractClient
from finorch.transport.exceptions import TransportStartJobException
from finorch.utils.cd import cd
from finorch.utils.job_status import JobStatus


SLURM_SCRIPT = """#!/bin/bash
#SBATCH --time=01:00:00
#SBATCH --mem=16G
#SBATCH --nodes=1
#SBATCH --ntasks-per-node=1

. .env
{python} -m finorch.wrapper.wrapper ozstar
"""


class OzStarClient(AbstractClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _write_environment(self, environment_file):
        with open(environment_file, "w") as f:
            for k, v in os.environ.items():
                if "()" in k:
                    continue
                else:
                    f.write(f'{k}="{v}"\n')

    def _submit_slurm_job(self, job_identifier, katscript):
        # Create the working directory for the job
        exec_dir = self._exec_path / job_identifier
        os.makedirs(exec_dir, exist_ok=True)

        with cd(exec_dir):
            # Write the katscript
            with open(exec_dir / 'script.k', 'w') as f:
                f.write(katscript)

            # Write the environment file
            self._write_environment(exec_dir / '.env')

            # Write the slurm submission script
            slurm_script_path = exec_dir / 'submit.sh'
            with open(slurm_script_path, 'w') as f:
                script = SLURM_SCRIPT.format(python=sys.executable)
                f.write(script)

            # Construct the sbatch command
            command = f"sbatch {slurm_script_path}"

            # Execute the sbatch command
            stdout = None
            try:
                stdout = subprocess.check_output(command, shell=True)
            except Exception:
                # Record the command and the output
                logging.error("Error: Command `{}` returned `{}`".format(command, stdout))
                raise TransportStartJobException("Unable to submit slurm job")

            # Record the command and the output
            logging.info("Success: Command `{}` returned `{}`".format(command, stdout))

            # Get the slurm id from the output
            try:
                return int(stdout.strip().split()[-1])
            except Exception:
                raise TransportStartJobException("Unable to submit slurm job")

    def _cancel_slurm_job(self, job_id):
        logging.info("Trying to terminate job {}...".format(job_id))

        # Construct the command
        command = "scancel {}".format(job_id)

        # Cancel the job
        stdout = subprocess.check_output(command, shell=True)

        # todo: Handle errors
        # Get the output
        logging.info("Command `{}` returned `{}`".format(command, stdout))

    def start_job(self, katscript):
        job_identifier = str(uuid.uuid4())

        logging.info("Starting job with the following script")
        logging.info(katscript)
        logging.info(job_identifier)

        slurm_id = self._submit_slurm_job(job_identifier, katscript)
        self.db.add_job(job_identifier, slurm_id)

        return job_identifier

    def terminate(self):
        return super().terminate()

    def get_jobs(self):
        jobs = self.db.get_jobs()
        return jobs

    def get_job_status(self, job_identifier):
        status = self.db.get_job_status(job_identifier)

        # If the job status is less than or equal to RUNNING, then we need to derive the current job status and update
        # the job status accordingly.
        new_status = status
        if status <= JobStatus.RUNNING:
            # Check if the job is completed, or started, or queued
            p = Path(self._exec_path)
            if (p / job_identifier / 'finished').exists():
                new_status = JobStatus.COMPLETED
            elif (p / job_identifier / 'started').exists():
                new_status = JobStatus.RUNNING
            else:
                new_status = JobStatus.QUEUED

        # Update the job if the status has changed
        if new_status != status:
            self.db.update_job_status(job_identifier, new_status)

        return new_status

    def get_job_file(self, job_identifier, file_path):
        full_file_path = Path(self._exec_path / job_identifier / file_path)

        if full_file_path.exists():
            try:
                with open(full_file_path, 'rb') as f:
                    return f.read()
            except Exception:
                return None, f"Unable to retrieve file {full_file_path} as the file could not be read."

        return None, f"Unable to retrieve file {full_file_path} as the file does not exist."

    def get_job_file_list(self, job_identifier):
        full_path = Path(self._exec_path / job_identifier)
        if full_path.exists():
            # list the files
            file_list = list()

            for p in full_path.rglob('*.*'):
                file_list.append([p.name, str(p), p.stat().st_size])

            return file_list

        return None, f"Unable to retrieve file list for the job identifier {job_identifier}"

    def stop_job(self, job_identifier):
        # If the current job status is less than running, then cancel the job
        if self.get_job_status(job_identifier) < JobStatus.RUNNING:
            # Tell slurm to cancel the job
            self._cancel_slurm_job(self.db.get_job_batch_id(job_identifier))

            # Mark the job as cancelled
            self.db.update_job_status(job_identifier, JobStatus.CANCELLED)



import socket


def test_port_open(port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # connect_ex returns 0 if the connection was successful
    result = sock.connect_ex(("127.0.0.1", int(port))) == 0
    sock.close()
    return result



import abc

from finorch.transport.exceptions import TransportConnectionException


class AbstractTransport(abc.ABC):
    """
    This is the transport class that should be inherited from to build various transport types
    """

    def __init__(self, session, exec_path, *args, **kwargs):
        """
        Any transport setup should be done in this function

        :param kwargs: Any additional parameters that are required to initialise the transport
        :return: None
        """
        self._session = session
        self._exec_path = exec_path
        self._connected = False
        self._port = None

    @property
    def exec_path(self):
        return self._exec_path

    @abc.abstractmethod
    def connect(self):
        """
        Connects the transport

        Should raise a TransportConnectionException in the event of a problem

        :return: None
        """

        if self._connected:
            raise TransportConnectionException("Transport is already connected")

    @abc.abstractmethod
    def disconnect(self):
        """
        Disconnects the transport

        :return: None
        """

        if not self._connected:
            raise TransportConnectionException("Transport is not connected")

    @abc.abstractmethod
    def start_job(self, katscript):
        """
        Starts a job using this transport using the model defined by the provided katscript

        Should raise a TransportStartJobException in the event of a problem

        :param katscript: The katscript defining the model to run
        :return: UUID representing the remote identifier for the job
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_job_status(self, job_identifier):
        """
        Gets the job status for the provided job identifier

        Should raise a TransportGetJobStatusException in the event of a problem

        :param job_identifier: The UUID of the job to get the status of
        :return: a JobStatus indicating the status of the job
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def stop_job(self, job_identifier):
        """
        Stops a job using this transport with the provided job identifier

        Should raise a TransportStopJobException in the event of a problem

        :param job_identifier: The UUID of the job to stop
        :return: None
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_jobs(self):
        """
        Fetches all remote jobs using this transport

        Should raise a TransportGetJobsException in the event of a problem

        :return: A list of dicts representing the details of the remote jobs
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_job_file_list(self, job_identifier):
        """
        Retrieves the file list for the specified job identifier

        Should raise a TransportGetJobFileListException in the event of a problem

        :param job_identifier: the UUID of the job to fetch the file list for
        :return: A list of JobFile objects
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_job_file(self, job_identifier, file_path):
        """
        Retrieves the specified file for the specified job identifier

        Should raise a TransportGetJobFileException in the event of a problem

        :param job_identifier: The UUID of the job to fetch the specified file for
        :param file_path: The path to the file to download
        :return: A bytes object representing the file that was downloaded
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def terminate(self):
        """
        Stops the client and kills any finesse jobs that are running associated with the transport

        Should raise a TransportTerminateException in the event of a problem

        :return: None
        """
        raise NotImplementedError()



import abc
import logging
import logging.handlers
import pathlib
import sys
import traceback
import xmlrpc.client
from threading import Thread
from time import sleep

from finorch.config.config import WrapperConfigManager
from finorch.utils.port import test_port_open
from finorch.utils.xmlrpc import XMLRPCServer


class AbstractWrapper(abc.ABC):
    """
    In this context a wrapper is the code that surrounds a finesse process. It is responsible for responding to
    communication from the client
    """
    def __init__(self):
        self._xml_rpc_server = None

    def set_server(self, server):
        """
        Sets the XMLRPC server where required. This is then used by the terminate() command

        :param server: The xmlrpc server instance
        :return: None
        """

        self._xml_rpc_server = server

    def terminate(self):
        """
        Called to terminate the XMLRPC server

        :return: True if the server was terminated successfully, False otherwise
        """

        if self._xml_rpc_server:
            self._xml_rpc_server.terminate()

        return True

    def start(self):
        """
        Called by the wrapper to start the finesse thread

        :return: None
        """

        Thread(target=self._run).start()

    def _run(self):
        port = WrapperConfigManager().get_port()

        # Wait for the XMLRPC server to start
        while not test_port_open(port):
            sleep(0.1)

        # Touch the 'started' file
        pathlib.Path('started').touch()

        try:
            logging.info("Starting finesse job")

            try:
                self.run()
            except Exception as exc:
                # An exception occurred, log the exception to the log file
                logging.error("Error running finesse job")
                logging.error(type(exc))
                logging.error(exc.args)
                logging.error(exc)

                # And log the stack trace
                exc_type, exc_value, exc_traceback = sys.exc_info()
                lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
                exc_log = ''.join('!! ' + line for line in lines)
                logging.error(exc_log)

            logging.info("Finesse job completed")
        finally:
            # Touch the 'finished' file
            pathlib.Path('finished').touch()

            try:
                # Kill the rpc server (If it's running). The deal with this is that when testing, since the
                # entire wrapper is itself running in a thread, the _run thread is not torn down with the
                # xml rpc server when terminate is called - therefor the xmlrpc server is killed, but this
                # thread is still running. When this code is hit, it will raise an exception
                wrapper_rpc_client = xmlrpc.client.ServerProxy(
                    f'http://localhost:{port}/rpc',
                    allow_none=True
                )
                wrapper_rpc_client.terminate()
            except Exception:
                pass

    def run(self):
        pass

    @staticmethod
    def prepare_log_file():
        """
        Creates the log file and sets up logging parameters
        :return: None
        """
        # Get the log file name
        log_file_name = pathlib.Path.cwd() / 'wrapper.log'

        from importlib import reload
        logging.shutdown()
        reload(logging)

        # Create the logger
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)

        # Create the log handler
        handler = logging.handlers.RotatingFileHandler(log_file_name, maxBytes=10485760, backupCount=5)
        handler.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))

        # Add the handler to the logger
        logger.addHandler(handler)

    @staticmethod
    def start_wrapper(session_klass):
        """
        Starts the wrapper.

        :return: None
        """
        wrapper = session_klass.wrapper_klass()

        # Create the XMLRPC server on a random port
        with XMLRPCServer(('localhost', 0)) as server:
            server.register_introspection_functions()

            wrapper.set_server(server)
            server.register_instance(wrapper)

            # Save the port in the wrapper configuration and start finesse
            port = server.server_address[1]
            logging.info(f"Port is {port}")
            WrapperConfigManager().set_port(port)

            wrapper.start()

            # Run the server's main loop
            server.serve_forever()



import logging

from finorch.config.config import api_config_manager
from finorch.sessions.ozstar.client import OzStarClient
from finorch.sessions.abstract_session import AbstractSession
from finorch.sessions.ozstar.wrapper import OzStarWrapper
from finorch.transport.ssh import SshTransport


class OzStarSession(AbstractSession):
    callsign = "ozstar"
    client_klass = OzStarClient
    wrapper_klass = OzStarWrapper
    transport_klass = SshTransport

    def __init__(self, exec_path, username, python_path, env_file=None, *args, **kwargs):
        """
        Creates a new ozstar session that can be used to run finesse jobs in parallel on ozstar.

        :param exec_path: The path to where jobs should be executed (and results stored), if not specified the path
        will be a temporary directory that is cleaned up when the client is terminated.
        """
        super().__init__()

        self._transport = OzStarSession.transport_klass(
            self,
            exec_path,
            username=username,
            python_path=python_path,
            env_file=env_file,
            host="farnarkle1.hpc.swin.edu.au",
            callsign=self.callsign,
            *args,
            **kwargs
        )

        ozstar_config = api_config_manager.get_section('ozstar')

        remote_port = ozstar_config.get('remote_port', None) if ozstar_config else None

        if remote_port:
            logging.info("Attempting to reconnect remote client last seen on remote port " + str(remote_port))
        else:
            logging.info("Attempting to connect remote client")

        remote_port = self._transport.connect(
            remote_port=remote_port
        )

        logging.info("Remote client connected on port " + str(remote_port))

        api_config_manager.set('ozstar', 'remote_port', str(remote_port))

    @property
    def transport(self):
        return self._transport



import logging
import multiprocessing.pool
import os
import sys
import uuid

from pathlib import Path
from finorch.sessions.abstract_client import AbstractClient
from finorch.sessions.abstract_wrapper import AbstractWrapper
from finorch.utils.job_status import JobStatus


def _start_wrapper(_exec_path, _job_identifier, _session_klass, katscript):
    """
    Executed in another process to start the job
    :return: None
    """
    exec_dir = _exec_path / _job_identifier
    os.makedirs(exec_dir)
    os.chdir(exec_dir)

    sys.stdout = open(str(exec_dir / 'out.log'), "w")
    sys.stderr = open(str(exec_dir / 'out.err'), "w")

    with open(exec_dir / 'script.k', 'w') as f:
        f.write(katscript)

    AbstractWrapper.prepare_log_file()
    AbstractWrapper.start_wrapper(_session_klass)


class LocalClient(AbstractClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Will create a pool with os.cpu_count() processes
        self._pool = multiprocessing.pool.Pool()

    def start_job(self, katscript):
        job_identifier = str(uuid.uuid4())

        self.db.add_job(job_identifier)

        logging.info("Starting job with the following script")
        logging.info(katscript)
        logging.info(job_identifier)

        self._pool.apply_async(
            _start_wrapper, (
                self._exec_path,
                job_identifier,
                self._session_klass,
                katscript
            )
        )

        return job_identifier

    def terminate(self):
        return super().terminate()

    def get_jobs(self):
        jobs = self.db.get_jobs()
        return jobs

    def get_job_status(self, job_identifier):
        status = self.db.get_job_status(job_identifier)

        # If the job status is less than or equal to RUNNING, then we need to derive the current job status and update
        # the job status accordingly.
        new_status = status
        if status <= JobStatus.RUNNING:
            # Check if the job is completed, or started, or queued
            p = Path(self._exec_path)
            if (p / job_identifier / 'finished').exists():
                new_status = JobStatus.COMPLETED
            elif (p / job_identifier / 'started').exists():
                new_status = JobStatus.RUNNING
            else:
                new_status = JobStatus.QUEUED

        # Update the job if the status has changed
        if new_status != status:
            self.db.update_job_status(job_identifier, new_status)

        return new_status

    def get_job_file(self, job_identifier, file_path):
        full_file_path = Path(self._exec_path / job_identifier / file_path)

        if full_file_path.exists():
            try:
                with open(full_file_path, 'rb') as f:
                    return f.read()
            except Exception:
                return None, f"Unable to retrieve file {full_file_path} as the file could not be read."

        return None, f"Unable to retrieve file {full_file_path} as the file does not exist."

    def get_job_file_list(self, job_identifier):
        full_path = Path(self._exec_path / job_identifier)
        if full_path.exists():
            # list the files
            file_list = list()

            for p in full_path.rglob('*.*'):
                file_list.append([p.name, str(p), p.stat().st_size])

            return file_list

        return None, f"Unable to retrieve file list for the job identifier {job_identifier}"

    def stop_job(self, job_identifier):
        raise NotImplementedError()



import logging
from io import StringIO

import select
import xmlrpc.client
from socketserver import BaseRequestHandler, ThreadingTCPServer
from threading import Thread
from time import sleep

import paramiko

from finorch.config.config import api_config_manager
from finorch.transport.abstract_transport import AbstractTransport
from finorch.transport.exceptions import TransportConnectionException, TransportTerminateException, \
    TransportGetJobFileException, TransportGetJobFileListException, TransportGetJobStatusException, \
    TransportGetJobsException


class SshTransport(AbstractTransport):
    def __init__(self, session, exec_path, *args, **kwargs):
        super().__init__(session, exec_path, *args, **kwargs)

        self._ssh_client = None
        self._ssh_transport = None
        self._ssh_session = None

        self._host = kwargs['host']
        self._username = kwargs['username']
        self._ssh_password = kwargs.get('password', None)

        self._python_path = kwargs['python_path']
        self._env_file = kwargs.get('env_file', None)
        self._callsign = kwargs['callsign']
        self._ssh_port = kwargs.get('ssh_port', 22)

        self._remote_port = None

    def connect(self, *args, **kwargs):
        self._remote_port = kwargs['remote_port']
        self._remote_port = int(self._remote_port) if self._remote_port else None

        self._ssh_client = paramiko.SSHClient()
        self._ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        # If no ssh password is provided, then try to use the key from the api settings
        if not self._ssh_password:
            # Get the configuration for the specified session
            section = api_config_manager.get_section(self._callsign)
            if not section:
                raise TransportConnectionException("No password provided, and no key set")

            # Check if it has a key or not
            key = section.get('key', None)
            if not key:
                raise TransportConnectionException("No password provided, and no key set")

            skey = StringIO(key)
            pkey = paramiko.RSAKey.from_private_key(skey)

            # Set up a connection to the remote server using username and key
            self._ssh_client.connect(
                hostname=self._host,
                port=self._ssh_port,
                username=self._username,
                pkey=pkey,
            )
        else:
            # Set up a connection to the remote server using username and password
            self._ssh_client.connect(
                hostname=self._host,
                port=self._ssh_port,
                username=self._username,
                password=self._ssh_password
            )

        # Get the transport used by the client
        self._ssh_transport = self._ssh_client.get_transport()

        # First try to reconnect the previous port if it's set
        if self._remote_port:
            try:
                # Set up the ssh port forwarding
                self._forward_tunnel()

                # Connect the client rpc
                self._client_rpc = xmlrpc.client.ServerProxy(
                    f'http://localhost:{self._port}/rpc',
                    allow_none=True,
                    use_builtin_types=True
                )

                self._client_rpc.system.listMethods()

                # We're connected
                self._connected = True

                return self._remote_port
            except Exception:
                # Remote client is dead or invalid
                self._connection.server_close()
                self._connected = False

        # Remote client isn't running, start the remote client
        session = self._ssh_transport.open_channel("session")

        # Always try to make the execution directory and change to it
        command = f"bash --login -c \"mkdir -p {self.exec_path} && cd {self.exec_path} && "
        if self._env_file:
            command += f"source {self._env_file} && "

        command += f"{self._python_path} -m finorch.client.client {self._callsign}\""

        # Run the command to start the remote client
        logging.info(f"Executing command: {command}")
        session.exec_command(command)

        # Wait for the connection to close
        stdout, stderr = b'', b''
        while True:  # monitoring process
            # Reading from output streams
            while session.recv_ready():
                stdout += session.recv(1000)
            while session.recv_stderr_ready():
                stderr += session.recv_stderr(1000)
            if session.exit_status_ready():  # If completed
                break
            sleep(0.1)

        stdout = stdout.decode('utf-8')
        stderr = stderr.decode('utf-8')

        # Check that the command finished successfully
        if session.exit_status:
            raise TransportConnectionException(
                f"Unable to start remote server.\nstdout:\n{stdout}\n\nstderr:\n{stderr}\n"
            )

        # Finished with the session now
        session.close()

        # Parse the remote stdout
        stdout = stdout.splitlines()

        # Check if the client started successfully
        if stdout[0] == "error":
            # Report the error from the client
            raise TransportConnectionException('\n'.join(stdout[1:]))

        # Try to parse the first line of the output from the client as the port it is running on
        try:
            self._remote_port = int(stdout[0])
        except ValueError:
            raise TransportConnectionException(f"Unable to parse the port. Got {stdout[0]}")

        # Set up the ssh port forwarding
        self._forward_tunnel()

        # Connect the client rpc
        self._client_rpc = xmlrpc.client.ServerProxy(
            f'http://localhost:{self._port}/rpc',
            allow_none=True,
            use_builtin_types=True
        )

        self._client_rpc.set_exec_path(self.exec_path)

        # We're connected
        self._connected = True

        return self._remote_port

    def disconnect(self):
        raise NotImplementedError()

    def get_job_file(self, job_identifier, file_path):
        status = self._client_rpc.get_job_file(job_identifier, file_path)
        if type(status) is bytes:
            return status
        else:
            raise TransportGetJobFileException(status[1])

    def get_job_file_list(self, job_identifier):
        status = self._client_rpc.get_job_file_list(job_identifier)
        if type(status) is list:
            return status
        else:
            raise TransportGetJobFileListException(status[1])

    def get_job_status(self, job_identifier):
        status = self._client_rpc.get_job_status(job_identifier)
        if type(status) is int:
            return status
        else:
            raise TransportGetJobStatusException(status[1])

    def get_jobs(self):
        status = self._client_rpc.get_jobs()
        if type(status) is list:
            return status
        else:
            raise TransportGetJobsException(status[1])

    def start_job(self, katscript):
        return self._client_rpc.start_job(katscript)

    def stop_job(self, job_identifier):
        return self._client_rpc.stop_job(job_identifier)

    def terminate(self):
        print(self._connected)
        if not self._connected:
            raise TransportTerminateException("Client is not connected")

        self._client_rpc.terminate()
        self._connected = False

    """
    SSH supporting code. Adapted from the following:-
    https://github.com/paramiko/paramiko/blob/main/demos/forward.py
    https://stackoverflow.com/questions/11294919/port-forwarding-with-paramiko
    """

    class SshForwardServer(ThreadingTCPServer):
        daemon_threads = True
        allow_reuse_address = True

    class SshHandler(BaseRequestHandler):
        chain_host = None
        chain_port = None
        ssh_transport = None

        def handle(self):
            try:
                chan = self.ssh_transport.open_channel(
                    "direct-tcpip",
                    (self.chain_host, self.chain_port),
                    self.request.getpeername(),
                )
            except Exception as e:
                logging.error(
                    "Incoming request to %s:%d failed: %s"
                    % (self.chain_host, self.chain_port, repr(e))
                )
                return
            if chan is None:
                logging.error(
                    "Incoming request to %s:%d was rejected by the SSH server."
                    % (self.chain_host, self.chain_port)
                )
                return

            logging.info(
                "Connected!  Tunnel open %r -> %r -> %r"
                % (
                    self.request.getpeername(),
                    chan.getpeername(),
                    (self.chain_host, self.chain_port),
                )
            )
            while True:
                r, w, x = select.select([self.request, chan], [], [])
                if self.request in r:
                    data = self.request.recv(1024)
                    if len(data) == 0:
                        break
                    chan.send(data)
                if chan in r:
                    data = chan.recv(1024)
                    if len(data) == 0:
                        break
                    self.request.send(data)

            peername = self.request.getpeername()
            chan.close()
            self.request.close()
            logging.info("Tunnel closed from %r" % (peername,))

    def _forward_tunnel(self):
        # this is a little convoluted, but lets me configure things for the Handler
        # object.  (SocketServer doesn't give Handlers any way to access the outer
        # server normally.)
        class SubHander(SshTransport.SshHandler):
            chain_host = "localhost"
            chain_port = self._remote_port
            ssh_transport = self._ssh_transport

        self._connection = SshTransport.SshForwardServer(("localhost", 0), SubHander)

        # Get the local port
        self._port = self._connection.server_address[1]

        # Start a thread to run the server
        def server_thread():
            self._connection.serve_forever()

        Thread(target=server_thread, daemon=True).start()



import abc
from tempfile import NamedTemporaryFile

import finesse

from finorch.transport.exceptions import TransportGetJobSolutionException
from finorch.utils.job_status import JobStatus


class AbstractSession(abc.ABC):
    def __init__(self):
        self._transport = None

    def start_job(self, script):
        return self._transport.start_job(script)

    def stop_job(self, job_identifier):
        return self._transport.stop_job(job_identifier)

    def get_jobs(self):
        return self._transport.get_jobs()

    def get_job_status(self, job_identifier):
        return self._transport.get_job_status(job_identifier)

    def get_job_file(self, job_identifier, file_path):
        return self._transport.get_job_file(job_identifier, file_path)

    def get_job_file_list(self, job_identifier):
        return self._transport.get_job_file_list(job_identifier)

    def get_job_solution(self, job_identifier):
        if self.get_job_status(job_identifier) <= JobStatus.RUNNING:
            raise TransportGetJobSolutionException("Can't get solution as job is not yet finished")

        result = self._transport.get_job_file(job_identifier, 'data.pickle')

        with NamedTemporaryFile() as f:
            f.write(result)
            f.flush()
            return finesse.load(f.name, 'pickle')

    def terminate(self):
        self._transport.terminate()



from xmlrpc.server import SimpleXMLRPCRequestHandler, SimpleXMLRPCServer


class RequestHandler(SimpleXMLRPCRequestHandler):
    rpc_paths = ('/rpc',)


class XMLRPCServer(SimpleXMLRPCServer):
    def __init__(self, *args, **kwargs):
        super().__init__(requestHandler=RequestHandler, allow_none=True, *args, **kwargs)
        self._quit = False

    def serve_forever(self, **kwargs):
        """
        Overrides the serve_forever function to wait for the server to be ready to quit

        :param kwargs: N/A
        :return: None
        """
        while not self._quit:
            self.handle_request()

    def terminate(self):
        """
        Marks the server as ready for termination

        :return: None
        """

        self._quit = True



import abc
import atexit
import os
from pathlib import Path
from tempfile import TemporaryDirectory

from finorch.sessions.database import Database


class DatabaseNotConfiguredException(Exception):
    pass


class AbstractClient(abc.ABC):
    def __init__(self, session_klass):
        self._exec_path = None
        self._xml_rpc_server = None
        self._session_klass = session_klass
        self._db = None

    def set_server(self, server):
        """
        Sets the XMLRPC server where required. This is then used by the terminate() command

        :param server: The xmlrpc server instance
        :return: None
        """

        self._xml_rpc_server = server

    def set_exec_path(self, path):
        if path:
            # Path is already defined, so set up the path and make sure the directory is created
            self._exec_path = Path(path).absolute()
            os.makedirs(self._exec_path, exist_ok=True)
        else:
            # If the path is not specified, create a temporary directory and mark it for cleanup when the client exits
            tmpdir = TemporaryDirectory()
            self._exec_path = Path(tmpdir.name)
            atexit.register(lambda: tmpdir.cleanup())

        self._db = Database(self._exec_path)

    @property
    def db(self):
        if not self._db:
            raise DatabaseNotConfiguredException()

        return self._db

    def terminate(self):
        """
        Called to terminate the XMLRPC server

        :return: True if the server was terminated successfully, False otherwise
        """

        if self._xml_rpc_server:
            self._xml_rpc_server.terminate()

        return True

    @abc.abstractmethod
    def start_job(self, katscript):
        raise NotImplementedError()

    @abc.abstractmethod
    def stop_job(self, job_identifier):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_jobs(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_job_status(self, job_identifier):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_job_file(self, job_identifier, file_path):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_job_file_list(self, job_identifier):
        raise NotImplementedError()



from pathlib import Path

import finesse

from finorch.sessions.abstract_wrapper import AbstractWrapper


class CitWrapper(AbstractWrapper):
    def run(self):
        katscript = open('script.k', 'r').read()

        kat = finesse.Model()
        kat.parse(katscript)
        out = kat.run()
        finesse.save(out, Path.cwd() / "data.pickle")



import logging
import sys
import traceback

from finorch.sessions import session_map
from finorch.sessions.abstract_wrapper import AbstractWrapper


def run():
    # Attempt to start the wrapper and if there is an error print error to stdout, and the stack trace to stderr.
    try:
        AbstractWrapper.prepare_log_file()

        if len(sys.argv) != 2:
            raise Exception("Incorrect number of parameters")

        if sys.argv[1] not in session_map:
            raise Exception(f"Session type {sys.argv[1]} does not exist.")

        # Get the wrapper from the provided session parameter
        session_klass = session_map[sys.argv[1]]

        AbstractWrapper.start_wrapper(session_klass)
    except Exception as exc:
        # An exception occurred, log the exception to the log file
        logging.error("Error starting wrapper")
        logging.error(type(exc))
        logging.error(exc.args)
        logging.error(exc)

        # And log the stack trace
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        exc_log = ''.join('!! ' + line for line in lines)
        logging.error(exc_log)

        # Log to stdout and stderr
        print("error", flush=True)
        print(exc_log, flush=True)
        print("=EOF=")


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['finorch',
 'finorch.client',
 'finorch.config',
 'finorch.sessions',
 'finorch.sessions.cit',
 'finorch.sessions.local',
 'finorch.sessions.ozstar',
 'finorch.transport',
 'finorch.utils',
 'finorch.wrapper']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.29,<2.0.0',
 'appdirs>=1.4.4,<2.0.0',
 'exc>=0.92,<0.93',
 'finesse>=3.0a2,<4.0',
 'paramiko>=2.9.2,<3.0.0']

entry_points = \
{'console_scripts': ['remove_ssh_key = '
                     'scripts.finorch_key_manager:remove_ssh_key',
                     'set_ssh_key = scripts.finorch_key_manager:set_ssh_key']}

setup_kwargs = {
    'name': 'finorch',
    'version': '0.1.7',
    'description': 'Finesse version 3 job orchestrator and manager. This package can be used to run parallel jobs on various remote platforms, as well as locally.',
    'long_description': None,
    'author': 'Lewis Lakerink',
    'author_email': 'llakerink@swin.edu.au',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
import logging.handlers
import os
import sys
import traceback
from tempfile import NamedTemporaryFile

from finorch.config.config import client_config_manager
from finorch.sessions import session_map, LocalSession
from finorch.utils.xmlrpc import XMLRPCServer


def daemonize():  # pragma: no cover
    """
    do the UNIX double-fork magic, see Stevens' "Advanced
    Programming in the UNIX Environment" for details (ISBN 0201563177)
    http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
    """
    try:
        pid = os.fork()
        if pid > 0:
            # exit first parent
            sys.exit(0)
    except OSError as e:
        sys.stderr.write("fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
        sys.exit(1)

    # decouple from parent environment
    # os.chdir("/")
    os.setsid()
    os.umask(0)

    # do second fork
    try:
        pid = os.fork()
        if pid > 0:
            # exit from second parent
            sys.exit(0)
    except OSError as e:
        sys.stderr.write("fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
        sys.exit(1)

    # redirect standard file descriptors
    sys.stdout.flush()
    sys.stderr.flush()
    si = open('/dev/null', 'r')
    so = open('/dev/null', 'ab+', 0)
    se = open('/dev/null', 'ab+', 0)
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())


def start_client():
    """
    Starts the client.

    :return: None
    """
    # Get the client from the provided session parameter
    if len(sys.argv) != 2:
        raise Exception("Incorrect number of parameters")

    if sys.argv[1] not in session_map:
        raise Exception(f"Session type {sys.argv[1]} does not exist.")

    session_klass = session_map[sys.argv[1]]
    client = session_klass.client_klass(session_klass)

    # Create the XMLRPC server on a random port
    with XMLRPCServer(('localhost', 0)) as server:
        server.register_introspection_functions()

        client.set_server(server)
        server.register_instance(client)

        # Save the port in the client configuration
        port = server.server_address[1]
        client_config_manager.set_port(port)

        # Return the port via stdout to the caller
        print(server.server_address[1], flush=True)
        print("=EOF=", flush=True)

        n = NamedTemporaryFile()
        sys.stdout = open(n.name, "w")
        sys.stderr = sys.stdout

        if session_klass is not LocalSession:
            daemonize()

        # Run the server's main loop
        server.serve_forever()


def prepare_log_file():
    """
    Creates the log file and sets up logging parameters
    :return: None
    """
    # Reset any logging configuration
    from importlib import reload
    logging.shutdown()
    reload(logging)

    # Get the log file name
    log_file_name = client_config_manager.get_log_directory() / 'client.log'

    # Create the logger
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    # Create the log handler
    handler = logging.handlers.RotatingFileHandler(log_file_name, maxBytes=10485760, backupCount=5)
    handler.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))

    # Add the handler to the logger
    logger.addHandler(handler)


def run():
    # Attempt to start the client and if there is an error print error to stdout, and the stack trace to stderr.
    try:
        prepare_log_file()
        start_client()
    except Exception as exc:
        # An exception occurred, log the exception to the log file
        logging.error("Error starting client")
        logging.error(type(exc))
        logging.error(exc.args)
        logging.error(exc)

        # And log the stack trace
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        exc_log = ''.join('!! ' + line for line in lines)
        logging.error(exc_log)

        # Log to stdout and stderr
        print("error", flush=True)
        print(exc_log, flush=True)
        print("=EOF=")


if __name__ == '__main__':
    run()



from finorch.sessions.local.client import LocalClient
from finorch.sessions.abstract_session import AbstractSession
from finorch.sessions.local.wrapper import LocalWrapper
from finorch.transport.local import LocalTransport


class LocalSession(AbstractSession):
    callsign = "local"
    client_klass = LocalClient
    wrapper_klass = LocalWrapper
    transport_klass = LocalTransport

    def __init__(self, exec_path=None):
        """
        Creates a new local session that can be used to run finesse jobs in parallel locally.

        :param exec_path: The path to where jobs should be executed (and results stored), if not specified the path
        will be a temporary directory that is cleaned up when the client is terminated.
        """
        super().__init__()

        self._transport = LocalSession.transport_klass(self, exec_path)
        self._transport.connect()

    @property
    def transport(self):
        return self._transport



class TransportConnectionException(Exception):
    """
    Raised when a problem arises while connecting a transport
    """


class TransportStartJobException(Exception):
    """
    Raised when a problem arises while starting a job using a transport
    """


class TransportGetJobStatusException(Exception):
    """
    Raised when a problem arise while retrieving the status of a job
    """


class TransportGetJobSolutionException(Exception):
    """
    Raised when a problem arise while retrieving the solution for a job
    """


class TransportStopJobException(Exception):
    """
    Raised when a problem arises while stopping a job using a transport
    """


class TransportGetJobsException(Exception):
    """
    Raised when a problem arises while retrieving remote job information using a transport
    """


class TransportUpdateJobParametersException(Exception):
    """
    Raised when a problem arises while updating parameters of a remote job using a transport
    """


class TransportGetJobFileListException(Exception):
    """
    Raised when a problem arises while getting a job list of a remote job using a transport
    """


class TransportGetJobFileException(Exception):
    """
    Raised when a problem arises while downloading a job file of a remote job using a transport
    """


class TransportTerminateException(Exception):
    """
    Raised when a problem arise while terminating a client using a transport
    """



import os
import pathlib
from configparser import ConfigParser
from pathlib import Path

import appdirs

APPNAME = "finorch"
APPAUTHOR = "ADACS"


class _ConfigManager:
    """
    Used as a utility wrapper around the configuration ini files to make reading/writing configuration easier
    """

    def __init__(self, ini_file):
        self._ini_file = ini_file
        self._config = ConfigParser()

        # Read the configuration if possible
        self._read()

    @staticmethod
    def get_log_directory():
        """
        Gets the path to the log directory (And checks that the log directory exists)

        :return: A Path object representing the log directory.
        """
        p = Path(appdirs.user_config_dir(APPNAME, APPAUTHOR)) / 'logs'
        os.makedirs(p, exist_ok=True)
        return p

    @staticmethod
    def get_config_directory():
        """
        Gets the path to the config directory (And checks that the config directory exists)

        :return: A Path object representing the config directory.
        """
        p = Path(appdirs.user_config_dir(APPNAME, APPAUTHOR))
        os.makedirs(p, exist_ok=True)
        return p

    def get_section(self, section):
        """
        Read the specific section from the configuration file and return the configuration as a dict

        :param section: The section to return as a dict
        :return: The section (if it exists) as a dict, if the section does not exist it returns None
        """
        config_dict = {
            s: dict(self._config.items(s)) for s in self._config.sections()
        }

        return config_dict.get(section, None)

    def set(self, section, option, value):
        """
        Updates the configuration file with the specified section, option, and value

        :param section: The configuration section
        :param option: The configuration option
        :param value: The value to set
        :return: None
        """
        if not self._config.has_section(section):
            self._config.add_section(section)

        self._config.set(section, option, str(value))
        self._save()

    def _save(self):
        """
        Saves the configuration object to the configuration file

        :return: None
        """
        with open(self._ini_file, 'w') as f:
            self._config.write(f)
            f.flush()

    def _read(self):
        """
        Rereads the configuration file in to the configuration object

        :return: None
        """
        self._config.read(self._ini_file)


class _ApiConfigManager(_ConfigManager):
    """
    Configuration Manager for the API.
    """

    def __init__(self):
        super().__init__(self.get_config_directory() / "api.ini")


class _ClientConfigManager(_ConfigManager):
    """
    Configuration Manager for the Client.
    """

    def __init__(self):
        super().__init__(self.get_config_directory() / "client.ini")

    def get_port(self):
        """
        Gets the clients last port

        :return: The last port the client was running on, or None
        """
        self._read()

        if section := self.get_section("main"):
            return section.get("port", None)

        return None

    def set_port(self, port):
        """
        Sets the clients current port

        :return: None
        """
        self._read()
        self.set("main", "port", port)


class WrapperConfigManager(_ConfigManager):
    """
    Configuration Manager for the Wrapper.
    """

    def __init__(self, path=None):
        super().__init__((pathlib.Path(path) if path else pathlib.Path.cwd()) / "wrapper.ini")

    def get_port(self):
        """
        Gets the wrapper's port

        :return: The port the wrapper is running on, or None
        """
        self._read()

        if section := self.get_section("main"):
            return section.get("port", None)

        return None

    def set_port(self, port):
        """
        Sets the clients current port

        :return: None
        """
        self._read()
        self.set("main", "port", port)


# Create a config manager singleton to avoid issues with concurrency
api_config_manager = _ApiConfigManager()
client_config_manager = _ClientConfigManager()
